/*****************************************************************************
 * HASM64                                                                    *
 *****************************************************************************/

/*****************************************************************************
 * File name:     scan64.c                                                   *
 * Created:       2001-05-26 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2004-05-31 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2001-2004 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: scan64.c 20 2004-05-31 14:54:03Z hampa $ */


#include "hasm64.h"


static unsigned scn64_get_char (scanner_t *scn);


static unsigned keycnt = 0;

static keyword_t keywords[] = {
  { TOK_SYM_MODULE, "module" },
  { TOK_SYM_END, "end" },
  { TOK_SYM_ORG, "org" },
  { TOK_SYM_INCLUDE, "include" },
  { TOK_SYM_EQU, "=" },
  { TOK_SYM_NEQU, "!=" },
  { TOK_SYM_LESS, "<" },
  { TOK_SYM_GRTR, ">" },
  { TOK_SYM_LEQU, "<=" },
  { TOK_SYM_GEQU, ">=" },
  { TOK_SYM_OR, "|" },
  { TOK_SYM_AND, "&" },
  { TOK_SYM_XOR, "^" },
  { TOK_SYM_PLUS, "+" },
  { TOK_SYM_MINUS, "-" },
  { TOK_SYM_SHL, "<<" },
  { TOK_SYM_SHR, ">>" },
  { TOK_SYM_MUL, "*" },
  { TOK_SYM_DIV, "/" },
  { TOK_SYM_MOD, "mod" },
  { TOK_SYM_NOT, "~" },
  { TOK_SYM_LPAR, "(" },
  { TOK_SYM_RPAR, ")" },
  { TOK_SYM_COLON, ":" },
  { TOK_SYM_PERIOD, "." },
  { TOK_SYM_COMMA, "," },
  { TOK_SYM_LSPAR, "[", },
  { TOK_SYM_RSPAR, "]", },
  { TOK_SYM_LO, "lo" },
  { TOK_SYM_HI, "hi" },
  { TOK_SYM_IMM, "#" },
  { TOK_SYM_A, "a" },
  { TOK_SYM_X, "x" },
  { TOK_SYM_Y, "y" },
  { TOK_SYM_PC, "pc" },
  { TOK_OP_ADC, "adc" },
  { TOK_OP_AND, "and" },
  { TOK_OP_ASL, "asl" },
  { TOK_OP_BCC, "bcc" },
  { TOK_OP_BCS, "bcs" },
  { TOK_OP_BEQ, "beq" },
  { TOK_OP_BNE, "bne" },
  { TOK_OP_BPL, "bpl" },
  { TOK_OP_BMI, "bmi" },
  { TOK_OP_BVC, "bvc" },
  { TOK_OP_BVS, "bvs" },
  { TOK_OP_BRK, "brk" },
  { TOK_OP_BIT, "bit" },
  { TOK_OP_CLC, "clc" },
  { TOK_OP_CLD, "cld" },
  { TOK_OP_CLI, "cli" },
  { TOK_OP_CLV, "clv" },
  { TOK_OP_CMP, "cmp" },
  { TOK_OP_CPX, "cpx" },
  { TOK_OP_CPY, "cpy" },
  { TOK_OP_DEC, "dec" },
  { TOK_OP_DEX, "dex" },
  { TOK_OP_DEY, "dey" },
  { TOK_OP_EOR, "eor" },
  { TOK_OP_INC, "inc" },
  { TOK_OP_INX, "inx" },
  { TOK_OP_INY, "iny" },
  { TOK_OP_JMP, "jmp" },
  { TOK_OP_JSR, "jsr" },
  { TOK_OP_DB,  "db" },
  { TOK_OP_LDA, "lda" },
  { TOK_OP_LDX, "ldx" },
  { TOK_OP_LDY, "ldy" },
  { TOK_OP_LSR, "lsr" },
  { TOK_OP_NOP, "nop" },
  { TOK_OP_ORA, "ora" },
  { TOK_OP_PHA, "pha" },
  { TOK_OP_PHP, "php" },
  { TOK_OP_PLA, "pla" },
  { TOK_OP_PLP, "plp" },
  { TOK_OP_RB,  "rb" },
  { TOK_OP_ROL, "rol" },
  { TOK_OP_ROR, "ror" },
  { TOK_OP_RTI, "rti" },
  { TOK_OP_RTS, "rts" },
  { TOK_OP_SBC, "sbc" },
  { TOK_OP_SEC, "sec" },
  { TOK_OP_SED, "sed" },
  { TOK_OP_SEI, "sei" },
  { TOK_OP_STA, "sta" },
  { TOK_OP_STX, "stx" },
  { TOK_OP_STY, "sty" },
  { TOK_OP_TAX, "tax" },
  { TOK_OP_TAY, "tay" },
  { TOK_OP_TSX, "tsx" },
  { TOK_OP_TXA, "txa" },
  { TOK_OP_TXS, "txs" },
  { TOK_OP_TYA, "tya" },
  { TOK_OP_REP, "rep" },
  { 0, NULL }
};


/* Prepare the global list of keywords */
static
void scn64_prep_key (void)
{
  unsigned  i, j;
  keyword_t key;

  i = 1;
  while (keywords[i].str != NULL) {
    key = keywords[i];

    j = i;
    while ((j > 0) && (strcmp (key.str, keywords[j - 1].str) < 0)) {
      keywords[j] = keywords[j - 1];
      j -= 1;
    }

    keywords[j] = key;

    i += 1;
  }

  keycnt = i;
}

scanner_t *scn64_new (void)
{
  scanner_t   *scn;
  scan64_t    *s64;
  scn_class_t *cls;

  if (keycnt == 0) {
    scn64_prep_key ();
  }

  scn = scn_new ();
  if (scn == NULL) {
    return (NULL);
  }

  s64 = (scan64_t *) malloc (sizeof (scan64_t));
  if (s64 == NULL) {
    scn->del (scn);
    return (NULL);
  }

  s64->next = NULL;
  s64->fname = NULL;
  s64->fp = NULL;

  s64->buf_i = 0;
  s64->buf_n = 0;

  s64->keycnt = keycnt;
  s64->keywrd = keywords;
  scn->ext = s64;

  scn->del = &scn64_del;
  scn->get_char = &scn64_get_char;
  scn->scan = &scn64_scan;

  scn_set_clear (&scn->white);
  scn_set_add_el (&scn->white, 9);
  scn_set_add_el (&scn->white, 10);
  scn_set_add_el (&scn->white, 13);
  scn_set_add_el (&scn->white, 32);

  scn_set_clear (&scn->newline);
  scn_set_add_el (&scn->newline, 10);

  cls = scn_cls_new (TOK_IDENT);
  scn_set_add_els (&cls->start, 'a', 'z');
  scn_set_add_els (&cls->start, 'A', 'Z');
  scn_set_add_el (&cls->start, '_');
  scn_set_copy (&cls->run, &cls->start);
  scn_set_add_els (&cls->run, '0', '9');
  scn_class_add (scn, cls);

  cls = scn_cls_new (TOK_DIGIT);
  scn_set_add_els (&cls->start, '0', '9');
  scn_set_copy (&cls->run, &cls->start);
  scn_set_add_el (&cls->start, '$');
  scn_set_add_els (&cls->run, 'a', 'f');
  scn_set_add_els (&cls->run, 'A', 'F');
  scn_set_add_el (&cls->run, 'x');
  scn_set_add_el (&cls->run, 'b');
  scn_class_add (scn, cls);

  return (scn);
}

void scn64_del (scanner_t *scn)
{
  scan64_t *s64;

  if (scn == NULL) {
    return;
  }

  s64 = (scan64_t *) scn->ext;

  if (s64 == NULL) {
    scn_del (scn);
    return;
  }

  if (s64->next != NULL) {
    s64->next->del (s64->next);
  }

  if (s64->fname != NULL) {
    free (s64->fname);
  }

  if (s64->fp != NULL) {
    fclose (s64->fp);
  }

  free (s64);
  scn->ext = NULL;

  scn_del (scn);
}

int scn64_set_file (scanner_t *scn, char *fname)
{
  scan64_t *s64;

  s64 = (scan64_t *) scn->ext;

  if (s64->fname != NULL) {
    free (s64->fname);
  }

  if (s64->fp != NULL) {
    fclose (s64->fp);
  }

  s64->fname = str_clone (fname);

  s64->fp = fopen (fname, "rb");
  if (s64->fp == NULL) {
    return (1);
  }

  s64->buf_i = 0;
  s64->buf_n = 0;
  scn->curcnt = 0;

  return (0);
}

char *scn64_get_file (scanner_t *scn)
{
  scan64_t *s64;

  s64 = (scan64_t *) scn->ext;

  return (s64->fname);
}

void scn64_set_next (scanner_t *scn, scanner_t *next)
{
  scan64_t *s64;

  s64 = (scan64_t *) scn->ext;

  s64->next = next;
}

scanner_t *scn64_get_next (scanner_t *scn)
{
  scan64_t *s64;

  s64 = (scan64_t *) scn->ext;

  return (s64->next);
}

static
unsigned scn64_get_char (scanner_t *scn)
{
  scan64_t *s64;

  s64 = (scan64_t *)scn->ext;

  if (s64->fp == NULL) {
    return (SCN_EOF);
  }

  if (s64->buf_i >= s64->buf_n) {
    s64->buf_n = (unsigned) fread (s64->buf, 1, 1024, s64->fp);
    s64->buf_i = 0;
  }

  if (s64->buf_i < s64->buf_n) {
    return (s64->buf[s64->buf_i++]);
  }

  return (SCN_EOF);
}

void scn64_scan (scanner_t *scn)
{
  unsigned i, j, k;
  int      r;
  unsigned c;
  scan64_t *s64;

  s64 = (scan64_t *)scn->ext;

  scn_scan (scn);

  while (scn_tok_is_chr (scn, ';')) {
    c = scn_get_chr (scn, 0);
    while ((c != 10) && (c != 13)) {
      if (c == SCN_EOF) {
        scn->tok_id = SCN_TOK_NONE;
        scn->tok_str[0] = 0;
        return;
      }

      scn_rm_chr (scn, 1);
      c = scn_get_chr (scn, 0);
    }

    scn_scan (scn);
  }

  if (scn_tok_is_chr (scn, '<')) {
    c = scn_get_chr (scn, 0);
    if ((c == '=') || (c == '<')) {
      scn_rm_chr (scn, 1);
      scn->tok_str[1] = c;
      scn->tok_str[2] = 0;
    }
  }
  else if (scn_tok_is_chr (scn, '>')) {
    c = scn_get_chr (scn, 0);
    if ((c == '=') || (c == '>')) {
      scn_rm_chr (scn, 1);
      scn->tok_str[1] = c;
      scn->tok_str[2] = 0;
    }
  }

  if (scn_tok_is_chr (scn, '"')) {
    i = 0;
    c = scn_get_chr (scn, 0);

    while ((i < SCN_TOK_MAX) && (c != '"')) {
      scn->tok_str[i] = c;
      i += 1;

      scn_rm_chr (scn, 1);
      c = scn_get_chr (scn, 0);
      if (c == SCN_EOF) {
        scn->tok_id = SCN_TOK_NONE;
        scn->tok_str[0] = 0;
      }
    }

    scn_rm_chr (scn, 1);

    if (i >= SCN_TOK_MAX) {
      scn->tok_id = SCN_TOK_NONE;
      scn->tok_str[0] = 0;
      return;
    }
    else {
      scn->tok_id = TOK_STRING;
      scn->tok_str[i] = 0;
      return;
    }
  }

  if (scn->tok_id == TOK_DIGIT) {
    return;
  }

  i = 1;
  j = s64->keycnt;

  while (i <= j) {
    k = (i + j) / 2;
    r = strcmp (scn->tok_str, s64->keywrd[k - 1].str);

    if (r < 0) {
      j = k - 1;
    }
    else if (r > 0) {
      i = k + 1;
    }
    else {
      scn->tok_id = s64->keywrd[k - 1].tok;
      return;
    }
  }
}
